/*  
 *  CS 4391  Programming Assignment 1
 *  Robert Copeland
 *
 *  This program uses OpenGL to demonstrate smooth shading 
 *  applied to Wavefront polygon models.
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>       
#include <GL/glu.h>
#include <GL/glut.h>

void mousemove( int x, int y );
int ReadObject (char *filename);
void CalcCenter (float *Centroid);
void redraw (void);
float CalcXYExtent (float *Center);
void Cross (float *v1, float *v2, float *vout);
void Normalize (float *vec);
void CalcNormals (void);

#ifndef PI 
#define PI 3.1415925358979
#endif

/* Exit codes for error/success */
#ifndef EXIT_FAILURE
#define EXIT_FAILURE -1
#endif
#ifndef EXIT_SUCCESS
#define EXIT_SUCCESS 0
#endif

/* Parameters dealing with window / clip planes. */
#define FOV 45.0         
#define ASPECT_RATIO 1.0
#define CLIP_NEAR 1
#define CLIP_FAR 500.0

/* In order to keep this efficient and uncluttered, static 
   arrays will be employed.  Here are the max sizes allowed. */

/* Max total vertices in the object   */
#define MAX_VERTS      400000
/* Max number of vertices per polygon */
#define MAX_POLY_VERTS 10  
/* Max number of faces in the object  */
#define MAX_FACES      100000

typedef struct {
  float coords[3];                    /* World coordinates of vertex */
  float vnorm[3];                     /* Components of vertex normals */
} vertex;

typedef struct {
  int vindex[MAX_POLY_VERTS];           /* indices into vertex list       */
  int numpts;                           /* total number of valid vertices */
  float norm[3];                        /* components of face normals     */
} face;

static vertex VertexList[MAX_VERTS];    /* List of vertices (global) */
static face FaceList[MAX_FACES];        /* List of faces (global)    */

int NumFaces = 0;                       /* Total number of faces    */
int NumVerts = 0;                       /* Total number of vertices */

int main(int argc, char *argv[]) {
   
  float Center[3];                      /* Centroid of object */
  float ObjDist;                        /* Max distance from object
                                           origin to a vertex */
  float D;                              /* Distance where object uses
                                           approx 80% of viewport */

  float lightZeroPosition[3] = { 0, 12, 0 }; 
  float lightZeroColor[3] = { 1.0, 0.0, 0.0 };
  float lightOnePosition[3] = { 0, 0, 20 };
  float lightOneColor[3] = { 0.0, 0.0, 1.0 };
  float lightTwoPosition[3] = { 10, 0, 0 }; 
  float lightTwoColor[3] = { 0.0, 1.0, 0.0 };

 /* Window intialization (using GLUT) */
  glutInit(&argc, argv); 
  glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH); 
  glutCreateWindow(argv[0]);
  glutDisplayFunc(redraw);
  glutMotionFunc(mousemove);

  if (argc < 2) {
    fprintf (stderr, "USAGE: %s <objectfile>\n", argv[0]);
    return EXIT_FAILURE;
  }
  
  if (ReadObject (argv[1]) == -1) {
    return EXIT_FAILURE;
  }

  CalcNormals ();
  CalcCenter (Center);
  lightZeroPosition[0] += Center[0];
  lightZeroPosition[1] += Center[1];
  lightZeroPosition[2] += Center[2];
  lightOnePosition[0] += Center[0];
  lightOnePosition[1] += Center[1];
  lightOnePosition[2] += Center[2];
  lightTwoPosition[0] += Center[0];
  lightTwoPosition[1] += Center[1];
  lightTwoPosition[2] += Center[2];
  ObjDist = CalcXYExtent (Center);

/* To determine D, we use the following heuristic: 
 *
 * Since the FOV is 45 degrees, the angle between the
 * eye point and the origin of the object must be 22.5 
 * degrees.  The tangent of this angle is ViewportExtent / D.
 * Since we want ObjDist to be .8 * ViewportExtent, we have
 * D = (ObjDist/.8) / tan 22.5.
 */

  D = (ObjDist / .8) / (float) (tan (22.5*PI/180));
 

 /* Cull backfaces, use zbuffer, and lighting */
  glEnable(GL_CULL_FACE);
  glEnable(GL_DEPTH_TEST);
  glEnable(GL_LIGHTING);

 /* Generate camera matrix */
  glMatrixMode(GL_PROJECTION);
  gluPerspective(FOV, ASPECT_RATIO, CLIP_NEAR, CLIP_FAR);
  
  glMatrixMode(GL_MODELVIEW);
  gluLookAt(0.0, 0.0, D,  
    Center[0], Center[1], Center[2],
    0.0, 1.0, 0.0);       
  glPushMatrix();

 /* OpenGL calls to set up lighting */
  glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER, 1);
  glLightfv(GL_LIGHT0, GL_POSITION, lightZeroPosition);
  glLightfv(GL_LIGHT0, GL_DIFFUSE, lightZeroColor);
  glEnable(GL_LIGHT0);
  glLightfv(GL_LIGHT1, GL_POSITION, lightOnePosition);
  glLightfv(GL_LIGHT1, GL_DIFFUSE, lightOneColor);
  glEnable(GL_LIGHT1);
  glLightfv(GL_LIGHT2, GL_POSITION, lightTwoPosition);
  glLightfv(GL_LIGHT2, GL_DIFFUSE, lightTwoColor);
  glEnable(GL_LIGHT2);

 /* Start the event loop and get to work */
  glutMainLoop();

  return EXIT_SUCCESS;
}

int ReadObject (char *filename) {
/* int ReadObject (char *filename);
 *
 * Purpose: Reads a simplified Wavefront .OBJ file.
 *
 * Parameters: filename is the file to read.
 *
 * Output: VertexList and FaceList are filled with the
 *         proper values as read from the file on success.
 *         On error, -1 is returned.
 */

  #define MAXLINE 160
  FILE *fp;
  char *args;
  int i;
  char line[MAXLINE];

  fp = fopen( filename, "r" );
  if (!fp) {
    fprintf (stderr, "Error: Cannot open file %s\n", filename);
    return -1;
  }
  
  while (!feof(fp)) {
    if ( !fgets( line, MAXLINE, fp ) ) break;

    if ( line[0] == 'v' && isspace( line[1] )) { /* Vertex encountered */
      if (NumVerts > MAX_VERTS-1) {
        fprintf (stderr, "Error: MAX_VERTS exceeded in object %s\n", 
                 filename);
        fclose (fp);
        return -1;
      }
      /* Read vertices */
      sscanf( &line[2], "%f%f%f", &VertexList[NumVerts].coords[0],
                                  &VertexList[NumVerts].coords[1],
                                  &VertexList[NumVerts].coords[2] ); 
      NumVerts++;
    } else if ( line[0] == 'f' && isspace( line[1] )) { 

      /* Face list encountered */ 
      if (NumFaces > MAX_FACES-1) {
        fprintf (stderr, "Error: MAX_FACES exceeded in object %s\n", 
                 filename);
        fclose (fp);
        return -1;
      }
      FaceList[NumFaces].numpts = 0;
      
      args = strtok( &line[2], " \t\v\n" );
      while ( args ) {

        FaceList[ NumFaces ].vindex[ FaceList[ NumFaces ].numpts ] = 
          atoi( args ) - 1;

        FaceList[ NumFaces ].numpts++;
        if ( FaceList[ NumFaces ].numpts > MAX_POLY_VERTS) {
          fprintf (stderr, "Error: MAX_POLY_VERTS exceeded in object %s\n",
                   filename);
          fclose (fp);
          return -1;
        }
        args = strtok( NULL, " \t\v\n" );
      }
      NumFaces++;

    } else { 
      // fprintf( stderr, "Error: Unknown line %s, ignoring.\n", line );
    }
  }
  fclose (fp);
  return 0;
}

void CalcCenter (float *Centroid) {
/* float CalcCenter (float *Centroid);
 *
 * Purpose: Calculates the center of an object by averaging all the vertices.
 *
 * Parameters: Centroid points to the array of components to be updated.
 *
 * Output: The array pointed to by Centroid is filled with the average of all
 *         vertices.
 */
  int i,k;
  
  for ( i=0; i < NumVerts; i++ ) {
    for ( k=0; k < 3; k++ )
      Centroid[k] += VertexList[i].coords[k];
  }
  for ( k=0; k < 3; k++ ) 
    Centroid[k] /= (float) NumVerts;
}
    
float CalcXYExtent (float *Center) {
/* float CalcXYExtent (float *Center);
 *
 * Purpose: Calculates the largest distance between a vertex of an object
 *          and its origin in the XY plane.  This is used to provide a rough
 *          idea of how "large" the object will appear (note that the full
 *          distance formula is NOT used, because the viewing frustum is
 *          square, not round, and we want to ignore depth in this calc).
 *
 * Parameters: Center points to the origin of the object (as an array of 
 *             components).
 *
 * Output: The largest difference in the XY plane of a vertex and the origin
 *         of the object is returned.
 */

  int i,k;

  float Delta[] = { 0.0, 0.0 };
  float MaxDist = 0.0;  

  for ( i=0; i < NumVerts; i++ ) {
    for ( k=0; k < 2; k++ ) {
     /* Find the difference in both X and Y. */
      Delta[k] = VertexList[i].coords[k] - Center[k];
     /* Update if any Delta[k] is larger than MaxDist */
      MaxDist = (Delta[k] > MaxDist) ? Delta[k] : MaxDist;
    }
  }
  return MaxDist;
}

void CalcNormals (void) {
/* void CalcNormals (void);
 *
 * Purpose: Calculates the face / vertex normals from the object. FaceList 
 *          and VertexList must hold a valid object for this routine to have
 *          any value.
 * 
 * Parameters: None.
 *
 * Output:  All normals for the object are calculated.
 */

  int i,j,k;
  float vec1[3], vec2[3]; 

 /* Initialize vertex normals */
  for (i=0; i < NumVerts; i++) 
    VertexList[i].vnorm[0] = VertexList[i].vnorm[1] =  
    VertexList[i].vnorm[2] = 0.0;

  for (i=0; i<NumFaces; i++) {
   /* Calculate face normal */
    for (k=0; k < 3; k++ ) {
       vec1[k] = VertexList[FaceList[i].vindex[1]].coords[k] -
                 VertexList[FaceList[i].vindex[0]].coords[k];
       vec2[k] = VertexList[FaceList[i].vindex[2]].coords[k] -
                 VertexList[FaceList[i].vindex[1]].coords[k]; 
       Cross (vec1, vec2, FaceList[i].norm);
       Normalize (FaceList[i].norm);
    }
   /* Add normal to every vertex of the face */
    for (j=0; j < FaceList[i].numpts; j++ ) {
      for (k=0; k < 3; k++) 
        VertexList[FaceList[i].vindex[j]].vnorm[k]+=FaceList[i].norm[k];
    }
  }
 /* Normalize vertex normals */
  for (i=0; i < NumVerts; i++) {
    Normalize (VertexList[i].vnorm);
  }
}

void Cross (float *v1, float *v2, float *vout) {
/* void Cross (float *v1, float *v2, float *vout);
 *
 * Purpose : Computes the cross product of v1 and v2.
 *
 * Parameters : v1 and v2 are 3 component vectors (arrays) 
 *              to cross; vout points to the result array.
 * 
 * Output : vout is the un-normalized cross product of v1 and v2.
 */

  vout[0] = v1[1] * v2[2] - v2[1] * v1[2];
  vout[1] = - (v1[0] * v2[2] - v1[2] * v2[0]);
  vout[2] = v1[0] * v2[1] - v1[1] * v2[0];

}

void Normalize (float *vec) {
/* void Normalize (float *vec);
 *
 * Purpose : Divides vec by its magnitude.
 *
 * Parameters : vec is the 3 component vector to normalize.
 * 
 * Output : vec is divided by its magnitude.
 */

  int k;
  float mag;
  
  mag = sqrt (vec[0]*vec[0] + vec[1]*vec[1] + vec[2]*vec[2]);
  if (mag) {
    for (k = 0; k < 3; k++) 
      vec[k] /= mag;
  }
}

void DrawModel(void) {
/* void DrawModel (void)
 *
 * Purpose : Draws the object.
 *
 * Parameters : none.
 *
 * Output : Object is drawn.
 */
 int i,j;
 
 float *vnorm;
 float *vert;

 for (i=0; i < NumFaces; i++) { 

  /* draw a face */ 
   glBegin (GL_POLYGON);
   glColor3f( 1.0, 1.0, 1.0 );
   for (j=0; j < FaceList[i].numpts; j++) {
     vnorm = VertexList[FaceList[i].vindex[j]].vnorm;
     vert = VertexList[FaceList[i].vindex[j]].coords;
    /* Set vertex normal */
     glNormal3f ((GLfloat) *vnorm, (GLfloat) *(vnorm+1), (GLfloat) *(vnorm+2));
    /* Set vertex */
     glVertex3f((GLfloat) *vert, (GLfloat) *(vert+1), (GLfloat) *(vert+2));
   }
   glEnd ();
 }
}

void redraw (void) {
  glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  DrawModel();
  glutSwapBuffers();
  glFlush();
}

void mousemove( int x, int y ) {
  static int oldx = -1, oldy;
  static float zangle = 0, xangle = 0;

  if ( oldx != -1 ) {
    if ( x - oldx > 0 ) {
      zangle += 5.0;
    } else if ( x - oldx < 0 ) { 
      zangle -= 5.0;
    }

    if ( y - oldy > 0 ) {
      xangle += 5.0;
    } else if ( y - oldy < 0 ) { 
      xangle -= 5.0;
    }

    glMatrixMode( GL_MODELVIEW );
    glPopMatrix();
    glPushMatrix();
    glRotatef( zangle, 0, 1.0, 0 );
    glRotatef( xangle, 1.0, 0, 0 );
    glutPostRedisplay();
  }
  oldx = x; oldy = y;
}
